@@ -10,6 +10,8 @@
 #define DEBUG
 #include <osmocom/core/logging.h>
 
+#include <osmocom/mgcp_client/mgcp_client.h>
+#include <osmocom/mgcp_client/mgcp_client_pool.h>
 
 enum {
 	DMAIN,
@@ -19,8 +21,8 @@ enum {
 	DMGW,
 };
 
-#define LOGHNB(x, ss, lvl, fmt, args ...) \
-	LOGP(ss, lvl, "%s " fmt, hnb_context_name(x), ## args)
+#define LOGHNB(HNB_CTX, ss, lvl, fmt, args ...) \
+	LOGP(ss, lvl, "%s " fmt, hnb_context_name(HNB_CTX), ## args)
 
 enum hnb_ctrl_node {
 	CTRL_NODE_HNB = _LAST_CTRL_NODE,
@@ -84,6 +86,7 @@ struct hnbgw_cnlink {
 	struct llist_head map_list;
 };
 
+/* The lifecycle of the hnb_context object is the same as its conn */
 struct hnb_context {
 	/*! Entry in HNB-global list of HNB */
 	struct llist_head list;
@@ -101,8 +104,7 @@ struct hnb_context {
 	/*! SCTP stream ID for RUA */
 	uint16_t rua_stream;
 
-	/*! True if a HNB-REGISTER-REQ from this HNB has been accepted. Note that
-	 * this entire data structure is freed if the HNB sends HNB-DE-REGISTER-REQ. */
+	/*! True if a HNB-REGISTER-REQ from this HNB has been accepted. */
 	bool hnb_registered;
 
 	/* linked list of hnbgw_context_map */
@@ -136,6 +138,12 @@ struct hnb_gw {
 		bool log_prefix_hnb_id;
 		unsigned int max_sccp_cr_payload_len;
 		struct mgcp_client_conf *mgcp_client;
+		struct {
+			char *local_addr;
+			uint16_t local_port;
+			char *remote_addr;
+			uint16_t remote_port;
+		} pfcp;
 	} config;
 	/*! SCTP listen socket for incoming connections */
 	struct osmo_stream_srv_link *iuh;
@@ -154,7 +162,14 @@ struct hnb_gw {
 		struct osmo_sccp_addr iucs_remote_addr;
 		struct osmo_sccp_addr iups_remote_addr;
 	} sccp;
-	struct mgcp_client *mgcp_client;
+	/* MGW pool, also includes the single MGCP client as fallback if no
+	 * pool is configured. */
+	struct mgcp_client_pool *mgw_pool;
+
+	struct {
+		struct osmo_pfcp_endpoint *ep;
+		struct osmo_pfcp_cp_peer *cp_peer;
+	} pfcp;
 };
 
 extern void *talloc_asn1_ctx;
@@ -178,3 +193,13 @@ void hnbgw_vty_init(struct hnb_gw *gw, void *tall_ctx);
 int hnbgw_vty_go_parent(struct vty *vty);
 
 bool hnbgw_requires_empty_sccp_cr(struct hnb_gw *gw, unsigned int ranap_msg_len);
+
+/* Return true when the user configured GTP mapping to be enabled, by configuring a PFCP link to a UPF.
+ * Return false when the user configured to skip GTP mapping and RANAP PS RAB Requests/Responses should be passed thru
+ * 1:1.
+ * GTP mapping means that there are two GTP tunnels, one towards HNB and one towards CN, and we forward payloads between
+ * the two tunnels, mapping the TEIDs and GTP addresses. */
+static inline bool hnb_gw_is_gtp_mapping_enabled(const struct hnb_gw *gw)
+{
+	return gw->config.pfcp.remote_addr != NULL;
+}
