@@ -19,15 +19,20 @@
  *
  */
 
+#include "config.h"
+
 /* an expired mapping is destroyed  after 1..2 * EXPIRY_TIMER_SECS */
 #define EXPIRY_TIMER_SECS	23
 
 #include <osmocom/core/timer.h>
 
+#include <osmocom/sigtran/sccp_helpers.h>
+
 #include <osmocom/hnbgw/hnbgw.h>
 #include <osmocom/hnbgw/hnbgw_rua.h>
 #include <osmocom/hnbgw/context_map.h>
 #include <osmocom/hnbgw/mgw_fsm.h>
+#include <osmocom/hnbgw/ps_rab_ass_fsm.h>
 
 const struct value_string hnbgw_context_map_state_names[] = {
 	{MAP_S_NULL     , "not-initialized"},
@@ -55,8 +60,21 @@ static int alloc_cn_conn_id(struct hnbgw_cnlink *cn, uint32_t *id_out)
 	uint32_t i;
 	uint32_t id;
 
-	for (i = 0; i < 0xffffffff; i++) {
+	/* SUA: RFC3868 sec 3.10.4:
+	 *    The source reference number is a 4 octet long integer.
+	 *    This is allocated by the source SUA instance.
+	 * M3UA/SCCP: ITU-T Q.713 sec 3.3:
+	 *    The "source local reference" parameter field is a three-octet field containing a
+	 *    reference number which is generated and used by the local node to identify the
+	 *    connection section after the connection section is set up.
+	 *    The coding "all ones" is reserved for future use.
+	 * Hence, let's simply use 24 bit ids to fit all link types (excluding 0x00ffffff).
+	 */
+
+	for (i = 0; i < 0x00ffffff; i++) {
 		id = cn->next_conn_id++;
+		if (cn->next_conn_id == 0x00ffffff)
+			cn->next_conn_id = 0;
 		if (!cn_id_in_use(cn, id)) {
 			*id_out = id;
 			return 1;
@@ -102,6 +120,8 @@ context_map_alloc_by_hnb(struct hnb_context *hnb, uint32_t rua_ctx_id,
 	map->rua_ctx_id = rua_ctx_id;
 	map->is_ps = is_ps;
 	map->scu_conn_id = new_scu_conn_id;
+	INIT_LLIST_HEAD(&map->ps_rab_ass);
+	INIT_LLIST_HEAD(&map->ps_rabs);
 
 	/* put it into both lists */
 	llist_add_tail(&map->hnb_list, &hnb->map_list);
@@ -147,6 +167,8 @@ int context_map_send_cached_msg(struct hnbgw_context_map *map)
 
 void context_map_deactivate(struct hnbgw_context_map *map)
 {
+	LOG_MAP(map, DMAIN, LOGL_INFO, "Deactivating\n");
+
 	/* set the state to reserved. We still show up in the list and
 	 * avoid re-allocation of the context-id until we are cleaned up
 	 * by the context_map garbage collector timer */
@@ -154,12 +176,22 @@ void context_map_deactivate(struct hnbgw_context_map *map)
 	if (map->state != MAP_S_RESERVED2)
 		map->state = MAP_S_RESERVED1;
 
+	/* Is SCCP still active and needs to be disconnected ungracefully? */
+	if (map->scu_conn_active) {
+		osmo_sccp_tx_disconn(map->hnb_ctx->gw->sccp.cnlink->sccp_user, map->scu_conn_id, NULL, 0);
+		map->scu_conn_active = false;
+	}
+
 	/* a possibly still existing MGW FSM must be terminated when the context
 	 * map is deactivated. (this is a cornercase) */
 	if (map->mgw_fi) {
 		mgw_fsm_release(map);
 		OSMO_ASSERT(map->mgw_fi == NULL);
 	}
+
+#if ENABLE_PFCP
+	hnbgw_gtpmap_release(map);
+#endif
 }
 
 static struct osmo_timer_list context_map_tmr;
@@ -181,6 +213,7 @@ static void context_map_tmr_cb(void *data)
 		case MAP_S_RESERVED2:
 			/* second time we see this reserved
 			 * entry: remove it */
+			LOG_MAP(map, DMAIN, LOGL_INFO, "Deallocating\n");
 			map->state = MAP_S_NULL;
 			llist_del(&map->cn_list);
 			llist_del(&map->hnb_list);
