@@ -49,6 +49,7 @@
 #include <osmocom/ctrl/ports.h>
 #include <osmocom/vty/telnet_interface.h>
 #include <osmocom/vty/logging.h>
+#include <osmocom/vty/misc.h>
 #include <osmocom/vty/command.h>
 #include <osmocom/vty/ports.h>
 
@@ -61,10 +62,15 @@
 #include <osmocom/sigtran/protocol/m3ua.h>
 #include <osmocom/sigtran/sccp_sap.h>
 
+#if ENABLE_PFCP
+#include <osmocom/pfcp/pfcp_proto.h>
+#endif
+
 #include <osmocom/hnbgw/hnbgw.h>
 #include <osmocom/hnbgw/hnbgw_hnbap.h>
 #include <osmocom/hnbgw/hnbgw_rua.h>
 #include <osmocom/hnbgw/hnbgw_cn.h>
+#include <osmocom/hnbgw/hnbgw_pfcp.h>
 #include <osmocom/hnbgw/context_map.h>
 
 static const char * const osmo_hnbgw_copyright =
@@ -98,9 +104,14 @@ static struct hnb_gw *hnb_gw_create(void *ctx)
 
 	context_map_init(gw);
 
+	gw->mgw_pool = mgcp_client_pool_alloc(gw);
 	gw->config.mgcp_client = talloc_zero(tall_hnb_ctx, struct mgcp_client_conf);
 	mgcp_client_conf_init(gw->config.mgcp_client);
 
+#if ENABLE_PFCP
+	gw->config.pfcp.remote_port = OSMO_PFCP_PORT;
+#endif
+
 	return gw;
 }
 
@@ -229,30 +240,71 @@ void ue_context_free(struct ue_context *ue)
 static int hnb_read_cb(struct osmo_stream_srv *conn)
 {
 	struct hnb_context *hnb = osmo_stream_srv_get_data(conn);
+	struct osmo_fd *ofd = osmo_stream_srv_get_ofd(conn);
 	struct msgb *msg = msgb_alloc(IUH_MSGB_SIZE, "Iuh rx");
 	int rc;
 
 	if (!msg)
 		return -ENOMEM;
 
+	OSMO_ASSERT(hnb);
 	/* we store a reference to the HomeNodeB in the msg->dest for the
-	 * benefit of varoius downstream processing functions */
+	 * benefit of various downstream processing functions */
 	msg->dst = hnb;
 
 	rc = osmo_stream_srv_recv(conn, msg);
-	if (rc == -EAGAIN) {
-		/* Notification received */
-		msgb_free(msg);
-		return 0;
+	/* Notification received */
+	if (msgb_sctp_msg_flags(msg) & OSMO_STREAM_SCTP_MSG_FLAGS_NOTIFICATION) {
+		union sctp_notification *notif = (union sctp_notification *)msgb_data(msg);
+		rc = 0;
+		switch (notif->sn_header.sn_type) {
+		case SCTP_ASSOC_CHANGE:
+			switch (notif->sn_assoc_change.sac_state) {
+			case SCTP_COMM_LOST:
+				LOGHNB(hnb, DMAIN, LOGL_NOTICE,
+				       "sctp_recvmsg(%s) = SCTP_COMM_LOST, closing conn\n",
+				       osmo_sock_get_name2(ofd->fd));
+				osmo_stream_srv_destroy(conn);
+				rc = -EBADF;
+				break;
+			case SCTP_RESTART:
+				LOGHNB(hnb, DMAIN, LOGL_NOTICE, "HNB SCTP conn RESTARTed, marking as HNBAP-unregistered\n");
+				hnb->hnb_registered = false;
+				/* The tx queue may be quite full after an SCTP RESTART: (SYS#6113)
+				 * The link may have been flaky (a possible reason for the peer restarting the conn) and
+				 * hence the kernel socket Tx queue may be full (no ACKs coming back) and our own userspace
+				 * queue may contain plenty of oldish messages to be sent. Since the HNB will re-register after
+				 * this, we simply drop all those old messages: */
+				osmo_stream_srv_clear_tx_queue(conn);
+				break;
+			}
+			break;
+		case SCTP_SHUTDOWN_EVENT:
+			LOGHNB(hnb, DMAIN, LOGL_NOTICE,
+			       "sctp_recvmsg(%s) = SCTP_SHUTDOWN_EVENT, closing conn\n",
+			       osmo_sock_get_name2(ofd->fd));
+			osmo_stream_srv_destroy(conn);
+			rc = -EBADF;
+			break;
+		}
+		goto out;
+	} else if (rc == -EAGAIN) {
+		/* Older versions of osmo_stream_srv_recv() not supporting
+		 * msgb_sctp_msg_flags() may still return -EAGAIN when an sctp
+		 * notification is received. */
+		rc = 0;
+		goto out;
 	} else if (rc < 0) {
-		LOGHNB(hnb, DMAIN, LOGL_ERROR, "Error during sctp_recvmsg()\n");
-		/* FIXME: clean up after disappeared HNB */
-		hnb_context_release(hnb);
+		LOGHNB(hnb, DMAIN, LOGL_ERROR, "Error during sctp_recvmsg(%s)\n",
+		       osmo_sock_get_name2(ofd->fd));
+		osmo_stream_srv_destroy(conn);
+		rc = -EBADF;
 		goto out;
 	} else if (rc == 0) {
-		hnb_context_release(hnb);
-		rc = -1;
-
+		LOGHNB(hnb, DMAIN, LOGL_NOTICE, "Connection closed sctp_recvmsg(%s) = 0\n",
+		       osmo_sock_get_name2(ofd->fd));
+		osmo_stream_srv_destroy(conn);
+		rc = -EBADF;
 		goto out;
 	} else {
 		msgb_put(msg, rc);
@@ -264,6 +316,10 @@ static int hnb_read_cb(struct osmo_stream_srv *conn)
 		rc = hnbgw_hnbap_rx(hnb, msg);
 		break;
 	case IUH_PPI_RUA:
+		if (!hnb->hnb_registered) {
+			LOGHNB(hnb, DMAIN, LOGL_NOTICE, "Discarding RUA as HNB is not registered\n");
+			goto out;
+		}
 		hnb->rua_stream = msgb_sctp_stream(msg);
 		rc = hnbgw_rua_rx(hnb, msg);
 		break;
@@ -284,6 +340,21 @@ out:
 	return rc;
 }
 
+static int hnb_closed_cb(struct osmo_stream_srv *conn)
+{
+	struct hnb_context *hnb = osmo_stream_srv_get_data(conn);
+	if (!hnb)
+		return 0; /* hnb_context is being freed, nothing do be done */
+
+	/* hnb: conn became broken, let's release the associated hnb.
+	 * conn object is being freed after closed_cb(), so unassign it from hnb
+	 * if available to avoid it freeing it again: */
+	hnb->conn = NULL;
+	hnb_context_release(hnb);
+
+	return 0;
+}
+
 struct hnb_context *hnb_context_alloc(struct hnb_gw *gw, struct osmo_stream_srv_link *link, int new_fd)
 {
 	struct hnb_context *ctx;
@@ -294,7 +365,7 @@ struct hnb_context *hnb_context_alloc(struct hnb_gw *gw, struct osmo_stream_srv_
 	INIT_LLIST_HEAD(&ctx->map_list);
 
 	ctx->gw = gw;
-	ctx->conn = osmo_stream_srv_create(tall_hnb_ctx, link, new_fd, hnb_read_cb, NULL, ctx);
+	ctx->conn = osmo_stream_srv_create(tall_hnb_ctx, link, new_fd, hnb_read_cb, hnb_closed_cb, ctx);
 	if (!ctx->conn) {
 		LOGP(DMAIN, LOGL_INFO, "error while creating connection\n");
 		talloc_free(ctx);
@@ -328,6 +399,8 @@ void hnb_context_release(struct hnb_context *ctx)
 {
 	struct hnbgw_context_map *map, *map2;
 
+	LOGHNB(ctx, DMAIN, LOGL_INFO, "Releasing HNB context\n");
+
 	/* remove from the list of HNB contexts */
 	llist_del(&ctx->list);
 
@@ -342,7 +415,13 @@ void hnb_context_release(struct hnb_context *ctx)
 	}
 	ue_context_free_by_hnb(ctx->gw, ctx);
 
-	osmo_stream_srv_destroy(ctx->conn);
+	if (ctx->conn) { /* we own a conn, we must free it: */
+		LOGHNB(ctx, DMAIN, LOGL_INFO, "Closing HNB SCTP connection %s\n",
+		     osmo_sock_get_name2(osmo_stream_srv_get_ofd(ctx->conn)->fd));
+		/* Avoid our closed_cb calling hnb_context_release() again: */
+		osmo_stream_srv_set_data(ctx->conn, NULL);
+		osmo_stream_srv_destroy(ctx->conn);
+	} /* else: we are called from closed_cb, so conn is being freed separately */
 
 	talloc_free(ctx);
 }
@@ -358,6 +437,9 @@ static int accept_cb(struct osmo_stream_srv_link *srv, int fd)
 	struct hnb_gw *gw = osmo_stream_srv_link_get_data(srv);
 	struct hnb_context *ctx;
 
+	LOGP(DMAIN, LOGL_INFO, "New HNB SCTP connection %s\n",
+	     osmo_sock_get_name2(fd));
+
 	ctx = hnb_context_alloc(gw, srv, fd);
 	if (!ctx)
 		return -ENOMEM;
@@ -592,6 +674,41 @@ static int hnb_ctrl_node_lookup(void *data, vector vline, int *node_type, void *
 	return 1;
 }
 
+static int hnbgw_mgw_setup(void)
+{
+	struct mgcp_client *mgcp_client_single;
+	unsigned int pool_members_initalized;
+
+	/* Initialize MGW pool. This initalizes and connects all MGCP clients that are currently configured in
+	 * the pool. Adding additional MGCP clients to the pool is possible but the user has to configure and
+	 * (re)connect them manually from the VTY. */
+	pool_members_initalized = mgcp_client_pool_connect(g_hnb_gw->mgw_pool);
+	if (pool_members_initalized) {
+		LOGP(DMGW, LOGL_NOTICE,
+		     "MGW pool with %u pool members configured, (ignoring MGW configuration in VTY node 'mgcp').\n",
+		     pool_members_initalized);
+		return 0;
+	}
+
+	/* Initialize and connect a single MGCP client. This MGCP client will appear as the one and only pool
+	 * member if there is no MGW pool configured. */
+	LOGP(DMGW, LOGL_NOTICE, "No MGW pool configured, using MGW configuration in VTY node 'mgcp'\n");
+	mgcp_client_single = mgcp_client_init(tall_hnb_ctx, g_hnb_gw->config.mgcp_client);
+	if (!mgcp_client_single) {
+		LOGP(DMGW, LOGL_ERROR, "MGW (single) client initalization failed\n");
+		return -EINVAL;
+	}
+	if (mgcp_client_connect(mgcp_client_single)) {
+		LOGP(DMGW, LOGL_ERROR, "MGW (single) connect failed at (%s:%u)\n",
+		     g_hnb_gw->config.mgcp_client->remote_addr,
+		     g_hnb_gw->config.mgcp_client->remote_port);
+		return -EINVAL;
+	}
+	mgcp_client_pool_register_single(g_hnb_gw->mgw_pool, mgcp_client_single);
+
+	return 0;
+}
+
 int main(int argc, char **argv)
 {
 	struct osmo_stream_srv_link *srv;
@@ -614,6 +731,7 @@ int main(int argc, char **argv)
 		exit(1);
 	}
 
+	vty_info.tall_ctx = tall_hnb_ctx;
 	vty_info.copyright = osmo_hnbgw_copyright;
 	vty_init(&vty_info);
 
@@ -622,6 +740,7 @@ int main(int argc, char **argv)
 	hnbgw_vty_init(g_hnb_gw, tall_hnb_ctx);
 	ctrl_vty_init(tall_hnb_ctx);
 	logging_vty_add_cmds();
+	osmo_talloc_vty_add_cmds();
 
 	/* Handle options after vty_init(), for --version */
 	handle_options(argc, argv);
@@ -648,14 +767,14 @@ int main(int argc, char **argv)
 		log_set_log_level(osmo_stderr_target,
 				  hnbgw_cmdline_config.log_level);
 
-	rc = telnet_init_dynif(tall_hnb_ctx, g_hnb_gw, vty_get_bind_addr(), OSMO_VTY_PORT_HNBGW);
+	rc = telnet_init_default(tall_hnb_ctx, g_hnb_gw, OSMO_VTY_PORT_HNBGW);
 	if (rc < 0) {
 		perror("Error binding VTY port");
 		exit(1);
 	}
 
-	g_hnb_gw->ctrl = ctrl_interface_setup_dynip2(g_hnb_gw, ctrl_vty_get_bind_addr(), OSMO_CTRL_PORT_HNBGW,
-						     hnb_ctrl_node_lookup, _LAST_CTRL_NODE_HNB);
+	g_hnb_gw->ctrl = ctrl_interface_setup2(g_hnb_gw, OSMO_CTRL_PORT_HNBGW, hnb_ctrl_node_lookup,
+					       _LAST_CTRL_NODE_HNB);
 	if (!g_hnb_gw->ctrl) {
 		LOGP(DMAIN, LOGL_ERROR, "Failed to create CTRL interface on %s:%u\n",
 		     ctrl_vty_get_bind_addr(), OSMO_CTRL_PORT_HNBGW);
@@ -701,17 +820,13 @@ int main(int argc, char **argv)
 	g_hnb_gw->iuh = srv;
 
 	/* Initialize and connect MGCP client. */
-	g_hnb_gw->mgcp_client = mgcp_client_init(tall_hnb_ctx, g_hnb_gw->config.mgcp_client);
-	if (!g_hnb_gw->mgcp_client) {
-		LOGP(DMGW, LOGL_ERROR, "MGW client initalization failed\n");
+	if (hnbgw_mgw_setup() != 0)
 		return -EINVAL;
-	}
-	if (mgcp_client_connect(g_hnb_gw->mgcp_client)) {
-		LOGP(DMGW, LOGL_ERROR, "MGW connect failed at (%s:%u)\n",
-		     g_hnb_gw->config.mgcp_client->remote_addr,
-		     g_hnb_gw->config.mgcp_client->remote_port);
-		return -EINVAL;
-	}
+
+#if ENABLE_PFCP
+	/* If UPF is configured, set up PFCP socket and send Association Setup Request to UPF */
+	hnbgw_pfcp_init(g_hnb_gw);
+#endif
 
 	if (hnbgw_cmdline_config.daemonize) {
 		rc = osmo_daemonize();
