@@ -18,6 +18,8 @@
  *
  */
 
+#include "config.h"
+
 #include <arpa/inet.h>
 #include <errno.h>
 
@@ -29,12 +31,17 @@
 #include <osmocom/sigtran/sccp_sap.h>
 #include <osmocom/sigtran/sccp_helpers.h>
 
+#if ENABLE_PFCP
+#include <osmocom/pfcp/pfcp_cp_peer.h>
+#endif
+
 #include <osmocom/hnbgw/hnbgw.h>
 #include <osmocom/hnbgw/hnbgw_rua.h>
 #include <osmocom/ranap/ranap_ies_defs.h>
 #include <osmocom/ranap/ranap_msg_factory.h>
 #include <osmocom/hnbgw/context_map.h>
 #include <osmocom/hnbgw/mgw_fsm.h>
+#include <osmocom/hnbgw/ps_rab_ass_fsm.h>
 #include <osmocom/ranap/RANAP_ProcedureCode.h>
 #include <osmocom/ranap/ranap_common.h>
 #include <osmocom/ranap/ranap_common_ran.h>
@@ -330,6 +337,8 @@ static int handle_cn_conn_conf(struct hnbgw_cnlink *cnlink,
 			       struct osmo_prim_hdr *oph)
 {
 	struct osmo_ss7_instance *ss7 = osmo_sccp_get_ss7(cnlink->gw->sccp.client);
+	struct hnbgw_context_map *map;
+
 	LOGP(DMAIN, LOGL_DEBUG, "handle_cn_conn_conf() conn_id=%d, addrs: called=%s calling=%s responding=%s\n",
 	     param->conn_id,
 	     osmo_sccp_addr_to_str_c(OTC_SELECT, ss7, &param->called_addr),
@@ -339,9 +348,17 @@ static int handle_cn_conn_conf(struct hnbgw_cnlink *cnlink,
 	/* Nothing needs to happen for RUA, RUA towards the HNB doesn't seem to know any confirmations to its CONNECT
 	 * operation. */
 
+	map = context_map_by_cn(cnlink, param->conn_id);
+	if (!map)
+		return 0;
+
+	/* SCCP connection is confirmed. Mark conn as active, i.e. requires a DISCONNECT to clean up the SCCP
+	 * connection. */
+	map->scu_conn_active = true;
+
 	/* If our initial SCCP CR was sent without data payload, then the initial RANAP message is cached and waiting to
 	 * be sent as soon as the SCCP connection is confirmed. See if that is the case, send cached data. */
-	context_map_send_cached_msg(context_map_by_cn(cnlink, param->conn_id));
+	context_map_send_cached_msg(map);
 
 	return 0;
 }
@@ -365,8 +382,9 @@ static int handle_cn_data_ind(struct hnbgw_cnlink *cnlink,
 		return 0;
 	}
 
-	/* Intercept RAB Assignment Request, Setup MGW FSM */
+	/* Intercept RAB Assignment Request, to map RTP and GTP between access and core */
 	if (!map->is_ps) {
+		/* Circuit-Switched. Set up mapping of RTP ports via MGW */
 		message = talloc_zero(map, ranap_message);
 		rc = ranap_ran_rx_co_decode(map, message, msgb_l2(oph->msg), msgb_l2len(oph->msg));
 
@@ -385,6 +403,40 @@ static int handle_cn_data_ind(struct hnbgw_cnlink *cnlink,
 		}
 
 		talloc_free(message);
+#if ENABLE_PFCP
+	} else {
+		struct hnb_gw *hnb_gw = cnlink->gw;
+		/* Packet-Switched. Set up mapping of GTP ports via UPF */
+		message = talloc_zero(map, ranap_message);
+		rc = ranap_ran_rx_co_decode(map, message, msgb_l2(oph->msg), msgb_l2len(oph->msg));
+
+		if (rc == 0) {
+			switch (message->procedureCode) {
+
+			case RANAP_ProcedureCode_id_RAB_Assignment:
+				/* If a UPF is configured, handle the RAB Assignment via ps_rab_ass_fsm, and replace the
+				 * GTP F-TEIDs in the RAB Assignment message before passing it on to RUA. */
+				if (hnb_gw_is_gtp_mapping_enabled(hnb_gw)) {
+					LOGP(DMAIN, LOGL_DEBUG,
+					     "RAB Assignment: setting up GTP tunnel mapping via UPF %s\n",
+					     osmo_sockaddr_to_str_c(OTC_SELECT, &hnb_gw->pfcp.cp_peer->remote_addr));
+					return hnbgw_gtpmap_rx_rab_ass_req(map, oph, message);
+				}
+				/* If no UPF is configured, directly forward the message as-is (no GTP mapping). */
+				LOGP(DMAIN, LOGL_DEBUG, "RAB Assignment: no UPF configured, forwarding as-is\n");
+				break;
+
+			case RANAP_ProcedureCode_id_Iu_Release:
+				/* Any IU Release will terminate the MGW FSM, the message itsself is not passed to the
+				 * FSM code. It is just forwarded normally by the rua_tx_dt() call below. */
+				hnbgw_gtpmap_release(map);
+				break;
+			}
+			ranap_ran_rx_co_free(message);
+		}
+
+		talloc_free(message);
+#endif
 	}
 
 	return rua_tx_dt(map->hnb_ctx, map->is_ps, map->rua_ctx_id,
@@ -541,6 +593,8 @@ int hnbgw_cnlink_init(struct hnb_gw *gw, const char *stp_host, uint16_t stp_port
 		LOGP(DRANAP, LOGL_NOTICE, "No cs7 instance configured for IuCS nor IuPS,"
 		     " creating default instance\n");
 		ss7 = osmo_ss7_instance_find_or_create(gw, 0);
+		if (!ss7)
+			return -1;
 		ss7->cfg.primary_pc = (23 << 3) + 5;
 	}
 
