@@ -18,6 +18,7 @@
  *
  */
 
+#include "config.h"
 
 #include <osmocom/core/msgb.h>
 #include <osmocom/core/utils.h>
@@ -39,6 +40,7 @@
 #include <osmocom/hnbgw/context_map.h>
 #include <osmocom/hnbap/HNBAP_CN-DomainIndicator.h>
 #include <osmocom/hnbgw/mgw_fsm.h>
+#include <osmocom/hnbgw/ps_rab_ass_fsm.h>
 #include <osmocom/ranap/RANAP_ProcedureCode.h>
 #include <osmocom/ranap/ranap_common.h>
 #include <osmocom/ranap/ranap_common_cn.h>
@@ -250,6 +252,8 @@ int rua_to_scu(struct hnb_context *hnb,
 		prim->u.disconnect.conn_id = map->scu_conn_id;
 		prim->u.disconnect.cause = cause;
 		release_context_map = true;
+		/* Mark SCCP conn as gracefully disconnected */
+		map->scu_conn_active = false;
 		break;
 	case OSMO_SCU_PRIM_N_UNITDATA:
 		prim->u.unitdata.called_addr = *remote_addr;
@@ -273,20 +277,39 @@ int rua_to_scu(struct hnb_context *hnb,
 
 	/* If there is data, see if it is a RAB Assignment message where we need to change the user plane information,
 	 * for RTP mapping via MGW (soon also GTP mapping via UPF). */
-	if (data && len && map && !map->is_ps && !release_context_map) {
-		message = talloc_zero(map, ranap_message);
-		rc = ranap_cn_rx_co_decode(map, message, msgb_l2(prim->oph.msg), msgb_l2len(prim->oph.msg));
+	if (data && len && map && !release_context_map) {
+		if (!map->is_ps) {
+			message = talloc_zero(map, ranap_message);
+			rc = ranap_cn_rx_co_decode(map, message, msgb_l2(prim->oph.msg), msgb_l2len(prim->oph.msg));
 
-		if (rc == 0) {
-			switch (message->procedureCode) {
-			case RANAP_ProcedureCode_id_RAB_Assignment:
-				/* mgw_fsm_handle_rab_ass_resp() takes ownership of prim->oph and (ranap) message */
-				return mgw_fsm_handle_rab_ass_resp(map, &prim->oph, message);
+			if (rc == 0) {
+				switch (message->procedureCode) {
+				case RANAP_ProcedureCode_id_RAB_Assignment:
+					/* mgw_fsm_handle_rab_ass_resp() takes ownership of prim->oph and (ranap) message */
+					return mgw_fsm_handle_rab_ass_resp(map, &prim->oph, message);
+				}
+				ranap_cn_rx_co_free(message);
 			}
-			ranap_cn_rx_co_free(message);
-		}
 
-		talloc_free(message);
+			talloc_free(message);
+#if ENABLE_PFCP
+		} else if (hnb_gw_is_gtp_mapping_enabled(hnb->gw)) {
+			/* map->is_ps == true and PFCP is enabled in osmo-hnbgw.cfg */
+			message = talloc_zero(map, ranap_message);
+			rc = ranap_cn_rx_co_decode(map, message, msgb_l2(prim->oph.msg), msgb_l2len(prim->oph.msg));
+
+			if (rc == 0) {
+				switch (message->procedureCode) {
+				case RANAP_ProcedureCode_id_RAB_Assignment:
+					/* ps_rab_ass_fsm takes ownership of prim->oph and RANAP message */
+					return hnbgw_gtpmap_rx_rab_ass_resp(map, &prim->oph, message);
+				}
+				ranap_cn_rx_co_free(message);
+			}
+
+			talloc_free(message);
+#endif
+		}
 	}
 
 	rc = osmo_sccp_user_sap_down(cn->sccp_user, &prim->oph);
