@@ -18,6 +18,8 @@
  *
  */
 
+#include "config.h"
+
 #include <string.h>
 
 #include <osmocom/core/socket.h>
@@ -362,6 +364,51 @@ DEFUN(cfg_hnbgw_iups_remote_addr,
 	return CMD_SUCCESS;
 }
 
+#if ENABLE_PFCP
+
+static struct cmd_node pfcp_node = {
+	PFCP_NODE,
+	"%s(config-hnbgw-pfcp)# ",
+	1,
+};
+
+DEFUN(cfg_hnbgw_pfcp, cfg_hnbgw_pfcp_cmd,
+      "pfcp", "Configure PFCP for GTP tunnel mapping")
+{
+	vty->node = PFCP_NODE;
+	return CMD_SUCCESS;
+}
+
+DEFUN(cfg_pfcp_remote_addr, cfg_pfcp_remote_addr_cmd,
+      "remote-addr IP_ADDR",
+      "Remote UPF's listen IP address; where to send PFCP requests\n"
+      "IP address\n")
+{
+	osmo_talloc_replace_string(g_hnb_gw, &g_hnb_gw->config.pfcp.remote_addr, argv[0]);
+	LOGP(DLPFCP, LOGL_NOTICE, "%p cfg: pfcp remote-addr %s\n", g_hnb_gw, g_hnb_gw->config.pfcp.remote_addr);
+	return CMD_SUCCESS;
+}
+
+DEFUN(cfg_pfcp_local_addr, cfg_pfcp_local_addr_cmd,
+      "local-addr IP_ADDR",
+      "Local address for PFCP\n"
+      "IP address\n")
+{
+	osmo_talloc_replace_string(g_hnb_gw, &g_hnb_gw->config.pfcp.local_addr, argv[0]);
+	return CMD_SUCCESS;
+}
+
+DEFUN(cfg_pfcp_local_port, cfg_pfcp_local_port_cmd,
+      "local-port <1-65535>",
+      "Local port for PFCP\n"
+      "IP port\n")
+{
+	g_hnb_gw->config.pfcp.local_port = atoi(argv[0]);
+	return CMD_SUCCESS;
+}
+
+#endif /* ENABLE_PFCP */
+
 static int config_write_hnbgw(struct vty *vty)
 {
 	vty_out(vty, "hnbgw%s", VTY_NEWLINE);
@@ -369,6 +416,7 @@ static int config_write_hnbgw(struct vty *vty)
 		VTY_NEWLINE);
 	if (g_hnb_gw->config.max_sccp_cr_payload_len != 999999)
 		vty_out(vty, " sccp cr max-payload-len %u%s", g_hnb_gw->config.max_sccp_cr_payload_len, VTY_NEWLINE);
+	osmo_tdef_vty_groups_write(vty, " ");
 	return CMD_SUCCESS;
 }
 
@@ -417,13 +465,18 @@ static int config_write_hnbgw_iups(struct vty *vty)
 	return CMD_SUCCESS;
 }
 
-static int config_write_hnbgw_mgcp(struct vty *vty)
+#if ENABLE_PFCP
+static int config_write_hnbgw_pfcp(struct vty *vty)
 {
-	vty_out(vty, " mgcp%s", VTY_NEWLINE);
-	mgcp_client_config_write(vty, "  ");
+	vty_out(vty, " pfcp%s", VTY_NEWLINE);
+	if (g_hnb_gw->config.pfcp.local_addr)
+		vty_out(vty, "  local-addr %s%s", g_hnb_gw->config.pfcp.local_addr, VTY_NEWLINE);
+	if (g_hnb_gw->config.pfcp.remote_addr)
+		vty_out(vty, "  remote-addr %s%s", g_hnb_gw->config.pfcp.remote_addr, VTY_NEWLINE);
 
 	return CMD_SUCCESS;
 }
+#endif
 
 void hnbgw_vty_init(struct hnb_gw *gw, void *tall_ctx)
 {
@@ -461,8 +514,19 @@ void hnbgw_vty_init(struct hnb_gw *gw, void *tall_ctx)
 	install_element_ve(&show_talloc_cmd);
 
 	install_element(HNBGW_NODE, &cfg_hnbgw_mgcp_cmd);
-	install_node(&mgcp_node, config_write_hnbgw_mgcp);
-
+	/* Deprecated: Old MGCP config without pooling support in MSC node: */
+	install_node(&mgcp_node, NULL);
 	mgcp_client_vty_init(tall_hnb_ctx, MGCP_NODE, g_hnb_gw->config.mgcp_client);
+
+	mgcp_client_pool_vty_init(HNBGW_NODE, MGW_NODE, " ", g_hnb_gw->mgw_pool);
+
+#if ENABLE_PFCP
+	install_node(&pfcp_node, config_write_hnbgw_pfcp);
+	install_element(HNBGW_NODE, &cfg_hnbgw_pfcp_cmd);
+	install_element(PFCP_NODE, &cfg_pfcp_local_addr_cmd);
+	install_element(PFCP_NODE, &cfg_pfcp_local_port_cmd);
+	install_element(PFCP_NODE, &cfg_pfcp_remote_addr_cmd);
+#endif
+
 	osmo_tdef_vty_groups_init(HNBGW_NODE, hnbgw_tdef_group);
 }
