@@ -115,6 +115,7 @@ struct mgw_fsm_priv {
 	struct osmo_prim_hdr *ranap_rab_ass_resp_oph;
 
 	/* MGW context */
+	struct mgcp_client *mgcpc;
 	struct osmo_mgcpc_ep *mgcpc_ep;
 	struct osmo_mgcpc_ep_ci *mgcpc_ep_ci_hnb;
 	struct osmo_mgcpc_ep_ci *mgcpc_ep_ci_msc;
@@ -122,12 +123,7 @@ struct mgw_fsm_priv {
 	uint16_t msc_rtp_port;
 };
 
-static const struct osmo_tdef mgw_tdefs[] = {
-	{.T = -2427, .default_val = 5, .desc = "timeout for MGCP response from MGW" },
-	{ }
-};
-
-static const struct osmo_tdef_state_timeout mgw_fsm_timeouts[32] = {
+struct osmo_tdef_state_timeout mgw_fsm_timeouts[32] = {
 	[MGW_ST_CRCX_HNB] = {.T = -1001 },
 	[MGW_ST_ASSIGN] = {.T = -1002 },
 	[MGW_ST_MDCX_HNB] = {.T = -1003 },
@@ -177,9 +173,16 @@ static void mgw_fsm_crcx_hnb_onenter(struct osmo_fsm_inst *fi, uint32_t prev_sta
 	mgw_info.codecs[0] = CODEC_IUFP;
 	mgw_info.codecs_len = 1;
 
-	epname = mgcp_client_rtpbridge_wildcard(map->hnb_ctx->gw->mgcp_client);
+	mgw_fsm_priv->mgcpc = mgcp_client_pool_get(map->hnb_ctx->gw->mgw_pool);
+	if (!mgw_fsm_priv->mgcpc) {
+		LOGPFSML(fi, LOGL_ERROR,
+			 "cannot ensure MGW endpoint -- no MGW configured, check configuration!\n");
+		osmo_fsm_inst_state_chg(fi, MGW_ST_FAILURE, 0, 0);
+		return;
+	}
+	epname = mgcp_client_rtpbridge_wildcard(mgw_fsm_priv->mgcpc);
 	mgw_fsm_priv->mgcpc_ep =
-	    osmo_mgcpc_ep_alloc(fi, MGW_EV_MGCP_TERM, map->hnb_ctx->gw->mgcp_client, mgw_tdefs, fi->id, "%s", epname);
+	    osmo_mgcpc_ep_alloc(fi, MGW_EV_MGCP_TERM, mgw_fsm_priv->mgcpc, mgw_fsm_T_defs, fi->id, "%s", epname);
 	mgw_fsm_priv->mgcpc_ep_ci_hnb = osmo_mgcpc_ep_ci_add(mgw_fsm_priv->mgcpc_ep, "to-HNB");
 
 	osmo_mgcpc_ep_ci_request(mgw_fsm_priv->mgcpc_ep_ci_hnb, MGCP_VERB_CRCX, &mgw_info, fi, MGW_EV_MGCP_OK,
@@ -240,20 +243,20 @@ static void mgw_fsm_assign_onenter(struct osmo_fsm_inst *fi, uint32_t prev_state
 {
 	struct mgw_fsm_priv *mgw_fsm_priv = fi->priv;
 	struct hnbgw_context_map *map = mgw_fsm_priv->map;
-	uint8_t encoded[IUH_MSGB_SIZE];
 	RANAP_RAB_AssignmentRequestIEs_t *ies;
-	int rc;
+	struct msgb *msg;
 
 	ies = &mgw_fsm_priv->ranap_rab_ass_req_message->msg.raB_AssignmentRequestIEs;
-	rc = ranap_rab_ass_req_encode(encoded, sizeof(encoded), ies);
-	if (rc < 0) {
+	msg = ranap_rab_ass_req_encode(ies);
+	if (!msg) {
 		LOGPFSML(fi, LOGL_ERROR, "failed to re-encode RAB-AssignmentRequest message\n");
 		osmo_fsm_inst_state_chg(fi, MGW_ST_FAILURE, 0, 0);
 		return;
 	}
 
 	LOGPFSML(fi, LOGL_DEBUG, "forwarding modified RAB-AssignmentRequest to HNB\n");
-	rua_tx_dt(map->hnb_ctx, map->is_ps, map->rua_ctx_id, encoded, rc);
+	rua_tx_dt(map->hnb_ctx, map->is_ps, map->rua_ctx_id, msg->data, msg->len);
+	msgb_free(msg);
 }
 
 static void mgw_fsm_assign(struct osmo_fsm_inst *fi, uint32_t event, void *data)
@@ -497,6 +500,11 @@ static void mgw_fsm_allstate_action(struct osmo_fsm_inst *fi, uint32_t event, vo
 
 	switch (event) {
 	case MGW_EV_MGCP_TERM:
+		/* Put MGCP client back into MGW pool */
+		if (mgw_fsm_priv->mgcpc) {
+			mgcp_client_pool_put(mgw_fsm_priv->mgcpc);
+			mgw_fsm_priv->mgcpc = NULL;
+		}
 		mgw_fsm_priv->mgcpc_ep = NULL;
 		LOGPFSML(fi, LOGL_ERROR, "Media gateway failed\n");
 		osmo_fsm_inst_state_chg(fi, MGW_ST_FAILURE, 0, 0);
@@ -553,6 +561,10 @@ static void mgw_fsm_pre_term(struct osmo_fsm_inst *fi, enum osmo_fsm_term_cause
 	struct hnbgw_context_map *map = mgw_fsm_priv->map;
 
 	if (mgw_fsm_priv->mgcpc_ep) {
+		/* Put MGCP client back into MGW pool */
+		struct mgcp_client *mgcp_client = osmo_mgcpc_ep_client(mgw_fsm_priv->mgcpc_ep);
+		mgcp_client_pool_put(mgcp_client);
+
 		osmo_mgcpc_ep_clear(mgw_fsm_priv->mgcpc_ep);
 		mgw_fsm_priv->mgcpc_ep = NULL;
 	}
@@ -772,10 +784,8 @@ int mgw_fsm_handle_rab_ass_resp(struct hnbgw_context_map *map, struct osmo_prim_
  *  \returns 0 on success; negative on error. */
 int mgw_fsm_release(struct hnbgw_context_map *map)
 {
-	if (!map->mgw_fi) {
-		LOGP(DMGW, LOGL_ERROR, "mgw_fsm_release() rua_ctx_id=%d, no MGW fsm -- ignored!\n", map->rua_ctx_id);
+	if (!map->mgw_fi)
 		return -EINVAL;
-	}
 
 	osmo_fsm_inst_dispatch(map->mgw_fi, MGW_EV_RELEASE, NULL);
 	return 0;
